import numpy as np
from geopandas import GeoSeries
from geopandas.array import from_shapely
from hypothesis import HealthCheck, settings
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays
from scipy.spatial import Voronoi
from shapely import geometry as sg
from shapely.affinity import scale, translate
from shapely.ops import unary_union, polygonize

hyp_settings = settings(
    deadline=None,
    max_examples=500,
    suppress_health_check=[HealthCheck.too_slow],
)

coord = st.floats(
    allow_infinity=False, allow_nan=False, max_value=1000, min_value=-1000,
).map(lambda x: round(x, 15))

st_points = arrays(
    elements=st.floats(
        allow_infinity=False, allow_nan=False, max_value=60, min_value=-60
    ),
    shape=(100, 2),
    dtype='float64'
)


@st.composite
def st_point_array(draw, min_size=0, max_size=30, geoseries=False):
    n = draw(st.integers(min_size, max_size))
    points = []
    for i in range(n):
        x_mid = draw(st.floats(-50, 50))
        y_mid = draw(st.floats(-50, 50))
        point = (np.random.rand(2) - 0.5) * 5
        point[0] = point[0] + x_mid
        point[1] = point[1] + y_mid
        points.append(sg.Point(point))

    result = from_shapely(points)
    if geoseries:
        result = GeoSeries(result)
    return result


@st.composite
def st_multipoint_array(draw, min_size=0, max_size=30, geoseries=False):
    n = draw(st.integers(min_size, max_size))
    lines = []
    for i in range(n):
        num_points = draw(st.integers(1, 10))
        x_mid = draw(st.floats(-50, 50))
        y_mid = draw(st.floats(-50, 50))
        points = (np.random.rand(num_points, 2) - 0.5) * 5
        points[:, 0] = points[:, 0] + x_mid
        points[:, 1] = points[:, 1] + y_mid
        lines.append(sg.MultiPoint(points))

    result = from_shapely(lines)
    if geoseries:
        result = GeoSeries(result)
    return result


@st.composite
def st_line_array(draw, min_size=0, max_size=30, geoseries=False):
    n = draw(st.integers(min_size, max_size))
    lines = []
    for i in range(n):
        line_len = draw(st.integers(2, 10))
        x_mid = draw(st.floats(-50, 50))
        y_mid = draw(st.floats(-50, 50))
        points = np.cumsum(np.random.rand(line_len, 2) - 0.5, axis=0)
        points[:, 0] = points[:, 0] + x_mid
        points[:, 1] = points[:, 1] + y_mid
        lines.append(sg.LineString(points))

    result = from_shapely(lines)
    if geoseries:
        result = GeoSeries(result)
    return result


def get_unique_points(
        n, x_range=(0, 10), x_grid_dim=101, y_range=(0, 10), y_grid_dim=101
):
    """
    Get array of unique points, randomly drawn from a uniform grid
    """
    xs, ys = np.meshgrid(
        np.linspace(x_range[0], x_range[1], x_grid_dim),
        np.linspace(y_range[0], y_range[1], y_grid_dim),
    )
    points = np.stack([xs.flatten(), ys.flatten()], axis=1)
    selected_inds = np.random.choice(np.arange(points.shape[0]), n, replace=False)
    return points[selected_inds, :]


@st.composite
def st_ring_array(draw, min_size=3, max_size=30, geoseries=False):
    assert min_size >= 3
    n = draw(st.integers(min_size, max_size))
    rings = []
    for i in range(n):
        rings.append(sg.LinearRing(get_unique_points(n)))

    result = from_shapely(rings)
    if geoseries:
        result = GeoSeries(result)
    return result


@st.composite
def st_multiline_array(draw, min_size=0, max_size=5, geoseries=False):
    n = draw(st.integers(min_size, max_size))
    multilines = []
    for i in range(n):
        m = draw(st.integers(1, 5))
        lines = []
        for j in range(m):
            line_len = draw(st.integers(2, 3))
            x_mid = draw(st.floats(-50, 50))
            y_mid = draw(st.floats(-50, 50))
            points = np.cumsum(np.random.rand(line_len, 2) - 0.5, axis=0)
            points[:, 0] = points[:, 0] + x_mid
            points[:, 1] = points[:, 1] + y_mid
            lines.append(sg.LineString(points))
        multilines.append(sg.MultiLineString(lines))

    result = from_shapely(multilines)
    if geoseries:
        result = GeoSeries(result)
    return result


@st.composite
def st_polygon(draw, n=10, num_holes=None, xmid=0, ymid=0):
    # Handle defaults
    if num_holes is None:
        num_holes = draw(st.integers(0, 4))

    # Build outer shell
    tries = 50
    poly = None
    while tries > 0:
        tries -= 1
        points = (np.random.rand(n, 2) - 0.5) * 100
        points[:, 0] = points[:, 0] + xmid
        points[:, 1] = points[:, 1] + ymid

        vor = Voronoi(points)
        mls = sg.MultiLineString([
            vor.vertices[s] for s in vor.ridge_vertices if all(np.array(s) >= 0)
        ])

        poly = unary_union(list(polygonize(mls)))
        poly = poly.intersection(sg.box(-50, -50, 50, 50))
        if isinstance(poly, sg.Polygon) and not poly.is_empty:
            break

    if not isinstance(poly, sg.Polygon):
        raise ValueError("Failed to construct polygon")

    # Build holes
    remaining_holes = num_holes
    tries = 50
    while remaining_holes > 0 and tries > 0:
        tries -= 1
        points = (np.random.rand(n, 2) - 0.5) * 50
        points[:, 0] = points[:, 0] + xmid
        points[:, 1] = points[:, 1] + ymid
        vor = Voronoi(points)
        mls = sg.MultiLineString([
            vor.vertices[s] for s in vor.ridge_vertices if all(np.array(s) >= 0)
        ])
        hole_components = [p for p in polygonize(mls) if poly.contains(p)]
        if hole_components:
            hole = unary_union([p for p in polygonize(mls) if poly.contains(p)])
            if isinstance(hole, sg.MultiPolygon):
                hole = hole.geoms[0]

            new_poly = poly.difference(hole)
            if isinstance(new_poly, sg.Polygon):
                poly = new_poly
                remaining_holes -= 1

    return poly


@st.composite
def st_polygon_array(draw, min_size=0, max_size=5, geoseries=False):
    n = draw(st.integers(min_value=min_size, max_value=max_size))
    sg_polygons = [
        draw(st_polygon(xmid=draw(st.floats(-50, 50)), ymid=draw(st.floats(-50, 50))))
        for _ in range(n)
    ]

    result = from_shapely(sg_polygons)
    if geoseries:
        result = GeoSeries(result)
    return result


@st.composite
def st_multipolygon_array(draw, min_size=0, max_size=5, geoseries=False):
    n = draw(st.integers(min_value=min_size, max_value=max_size))
    sg_multipolygons = []
    for _ in range(n):
        xmid=draw(st.floats(-50, 50))
        ymid=draw(st.floats(-50, 50))
        polygon = draw(st_polygon(xmid=xmid, ymid=ymid))
        m = draw(st.integers(min_value=1, max_value=4))

        # populate polygons with m copies of polygon
        polygons = [polygon] * m

        # translate polygons so they don't overlap
        _, _, last_x1, last_y1 = polygons[0].bounds
        for j in range(1, m):
            polygon = scale(polygons[j], 0.8, 0.5)
            poly_x0, poly_y0, poly_x1, poly_y1 = polygon.bounds
            new_polygon = translate(
                polygon, yoff=last_y1 - poly_y0 + 1
            )
            _, _, last_x1, last_y1 = new_polygon.bounds
            polygons[j] = new_polygon

        sg_multipolygons.append(sg.MultiPolygon(polygons))

    result = from_shapely(sg_multipolygons)
    if geoseries:
        result = GeoSeries(result)
    return result


@st.composite
def st_bounds(draw, x_min=-60, y_min=-60, x_max=60, y_max=60, orient=False):
    x_float = st.floats(
        allow_infinity=False, allow_nan=False, min_value=x_min, max_value=x_max,
    )
    y_float = st.floats(
        allow_infinity=False, allow_nan=False, min_value=y_min, max_value=y_max,
    )
    # Generate x range
    x0, x1 = draw(x_float), draw(x_float)

    # Generate y range
    y0, y1 = draw(y_float), draw(y_float)

    if orient:
        if x1 < x0:
            x0, x1 = x1, x0
        if y1 < y0:
            y0, y1 = y1, y0

    return x0, y0, x1, y1



import pandas.tests.extension.base as eb
import pytest

from spatialpandas.geometry import LineArray, LineDtype


def test_equality():
    a = LineArray([[0, 1], [1, 2, 3, 4], None, [-1, -2], []], dtype='float64')
    assert all(a == a)
    assert all(a[1:-1] == a[[1, 2, 3]])
    assert not any(a[1:-1] == a[[2, 3, 1]])


# Pandas-provided extension array tests
# -------------------------------------
# See http://pandas-docs.github.io/pandas-docs-travis/extending.html
#
# Fixtures
@pytest.fixture
def dtype():
    """A fixture providing the ExtensionDtype to validate."""
    return LineDtype(subtype='float64')


@pytest.fixture
def data():
    """Length-100 array for this type.
        * data[0] and data[1] should both be non missing
        * data[0] and data[1] should not gbe equal
        """
    return LineArray(
        [[0, 1], [1, 2, 3, 4], None, [-1, -2], []]*20, dtype='float64')


@pytest.fixture
def data_repeated(data):
    """
    Generate many datasets.
    Parameters
    ----------
    data : fixture implementing `data`
    Returns
    -------
    Callable[[int], Generator]:
        A callable that takes a `count` argument and
        returns a generator yielding `count` datasets.
    """
    def gen(count):
        for _ in range(count):
            yield data
    return gen


@pytest.fixture
def data_missing():
    """Length-2 array with [NA, Valid]"""
    return LineArray([None, [-1, 0, 1, 2]], dtype='int64')


@pytest.fixture(params=['data', 'data_missing'])
def all_data(request, data, data_missing):
    """Parametrized fixture giving 'data' and 'data_missing'"""
    if request.param == 'data':
        return data
    elif request.param == 'data_missing':
        return data_missing


@pytest.fixture
def data_for_sorting():
    """Length-3 array with a known sort order.
    This should be three items [B, C, A] with
    A < B < C
    """
    return LineArray([[1, 0], [2, 0], [0, 0]])


@pytest.fixture
def data_missing_for_sorting():
    """Length-3 array with a known sort order.
    This should be three items [B, NA, A] with
    A < B and NA missing.
    """
    return LineArray([[1, 0], None, [0, 0]])


@pytest.fixture
def data_for_grouping():
    """Data for factorization, grouping, and unique tests.
    Expected to be like [B, B, NA, NA, A, A, B, C]
    Where A < B < C and NA is missing
    """
    return LineArray(
        [[1, 0], [1, 0], None, None, [0, 0], [0, 0], [1, 0], [2, 0]])


@pytest.fixture
def na_cmp():
    return lambda x, y: x is None and y is None


@pytest.fixture
def na_value():
    return None


@pytest.fixture
def groupby_apply_op():
    return lambda x: [1] * len(x)


@pytest.fixture
def fillna_method():
    return 'ffill'


@pytest.fixture(params=[True, False])
def as_frame(request):
    return request.param


@pytest.fixture(params=[True, False])
def as_series(request):
    return request.param


@pytest.fixture(params=[True, False])
def use_numpy(request):
    return request.param


@pytest.fixture(params=[None, lambda x: x])
def sort_by_key(request):
    """
    copied from pandas.conftest. Not sure why importing this module isn't enough
    to register fixture with pytest

    Simple fixture for testing keys in sorting methods.
    Tests None (no key) and the identity key.
    """
    return request.param


# Subclass BaseDtypeTests to run pandas-provided extension array test suite
class TestGeometryConstructors(eb.BaseConstructorsTests):
    pass


class TestGeometryDtype(eb.BaseDtypeTests):
    pass


class TestGeometryGetitem(eb.BaseGetitemTests):
    @pytest.mark.skip(reason="non-None fill value not supported")
    def test_take_non_na_fill_value(self):
        pass

    @pytest.mark.skip(reason="non-None fill value not supported")
    def test_reindex_non_na_fill_value(self, data_missing):
        pass

    @pytest.mark.skip("Cannot mask with a boolean indexer containing NA values")
    def test_getitem_boolean_na_treated_as_false(self, data):
        pass

    @pytest.mark.skip("Passing an invalid index type is not supported")
    def test_getitem_invalid(self, data):
        pass


class TestGeometryGroupby(eb.BaseGroupbyTests):
    @pytest.mark.skip(
        reason="The truth value of an array with more than one element is ambiguous."
    )
    def test_groupby_apply_identity(self):
        pass


class TestGeometryInterface(eb.BaseInterfaceTests):
    # # NotImplementedError: 'GeometryList' does not support __setitem__
    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_copy(self):
        pass

    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_view(self):
        pass

    @pytest.mark.skip(reason="contains not supported")
    def test_contains(self):
        pass


class TestGeometryMethods(eb.BaseMethodsTests):
    # # AttributeError: 'RaggedArray' object has no attribute 'value_counts'
    @pytest.mark.skip(reason="value_counts not supported")
    def test_value_counts(self):
        pass

    # Ragged array elements don't support binary operators
    @pytest.mark.skip(reason="ragged does not support <= on elements")
    def test_combine_le(self):
        pass

    @pytest.mark.skip(reason="ragged does not support + on elements")
    def test_combine_add(self):
        pass

    @pytest.mark.skip(reason="combine_first not supported")
    def test_combine_first(self):
        pass

    @pytest.mark.skip(reason="ragged does not support insert with an invalid scalar")
    def test_insert_invalid(self, data, invalid_scalar):
        pass

    @pytest.mark.skip(
        reason="Searchsorted seems not implemented for custom extension arrays"
    )
    def test_searchsorted(self):
        pass

    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_fillna_copy_frame(self, data):
        pass

    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_fillna_copy_series(self, data):
        pass

    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_shift_0_periods(self, data):
        pass

    @pytest.mark.skip(
        reason="value_counts not yet supported"
    )
    def test_value_counts_with_normalize(self, data):
        pass

    @pytest.mark.skip(reason="ragged does not support where on elements")
    def test_where_series(self):
        pass

class TestGeometryPrinting(eb.BasePrintingTests):
    pass


class TestGeometryMissing(eb.BaseMissingTests):
    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_fillna_frame(self):
        pass

    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_fillna_series(self):
        pass



class TestGeometryReshaping(eb.BaseReshapingTests):
    @pytest.mark.skip(reason="__setitem__ not supported")
    def test_ravel(self):
        pass

    @pytest.mark.skip(reason="transpose with numpy array elements seems not supported")
    def test_transpose(self):
        pass



import numpy as np
from dask.dataframe.extensions import make_array_nonempty
from pandas.core.dtypes.dtypes import register_extension_dtype

from ..geometry._algorithms.intersection import (point_intersects_polygon,
                                                 segment_intersects_point)
from ..geometry.base import GeometryDtype
from ..geometry.basefixed import GeometryFixed, GeometryFixedArray
from ..utils import ngpjit


@register_extension_dtype
class PointDtype(GeometryDtype):
    _geometry_name = 'point'
    @classmethod
    def construct_array_type(cls, *args):
        return PointArray


class Point(GeometryFixed):

    @classmethod
    def construct_array_type(cls):
        return PointArray

    @classmethod
    def _shapely_to_coordinates(cls, shape):
        import shapely.geometry as sg
        if isinstance(shape, sg.Point):
            # Single point
            return np.asarray(shape.coords).ravel()
        else:
            raise ValueError("""
Received invalid value of type {typ}. Must be an instance of Point,
or MultiPoint""".format(typ=type(shape).__name__))

    def to_shapely(self):
        """
        Convert to shapely shape

        Returns:
            shapely MultiPoint shape
        """
        import shapely.geometry as sg
        return sg.Point(self.flat_values)

    @classmethod
    def from_shapely(cls, shape):
        """
        Build a spatialpandas Point object from a shapely shape

        Args:
            shape: A shapely MultiPoint or Point shape

        Returns:
            spatialpandas MultiPoint
        """
        return super().from_shapely(shape)

    @property
    def x(self):
        return self.flat_values[0]

    @property
    def y(self):
        return self.flat_values[1]

    @property
    def length(self):
        return 0.0

    @property
    def area(self):
        return 0.0

    def intersects_bounds(self, bounds):
        x0, y0, x1, y1 = bounds
        if x1 < x0:
            x0, x1 = x1, x0
        if y1 < y0:
            y0, y1 = y1, y0
        outside = (np.isnan(self.x) or
                   self.x < x0 or self.x > x1 or
                   self.y < y0 or self.y > y1)

        return not outside

    def _intersects_point(self, point):
        return self.x == point.x and self.y == point.y

    def _intersects_multipoint(self, multipoint):
        flat = multipoint.flat_values
        return np.any((self.x == flat[0::2]) & (self.y == flat[1::2]))

    def _intersects_line(self, line):
        buffer_values = line.buffer_values
        offsets = line.buffer_inner_offsets

        for i in range(len(offsets) - 1):
            start = offsets[i]
            stop = offsets[i + 1]
            flat = buffer_values[start:stop]
            xs = flat[0::2]
            ys = flat[1::2]
            bounds = (min(xs), min(ys), max(xs), max(ys))

            # Check bounding box
            intersects_bounds = self.intersects_bounds(bounds)
            if not intersects_bounds:
                continue

            # Check if point exactly intersects vertex of line
            intersects_vert = np.any((self.x == flat[0::2]) & (self.y == flat[1::2]))
            if intersects_vert:
                return True

            # Check if point is exactly on one of the segments of the line
            bx = self.x
            by = self.y
            for j in range(len(xs) - 1):
                ax0 = xs[j]
                ay0 = ys[j]
                ax1 = xs[j + 1]
                ay1 = ys[j + 1]
                if segment_intersects_point(ax0, ay0, ax1, ay1, bx, by):
                    return True

        return False

    def _intersects_polygon(self, polygon):
        return point_intersects_polygon(
            self.x, self.y, polygon.buffer_values, polygon.buffer_inner_offsets
        )

    def intersects(self, shape):
        from . import Line, MultiLine, MultiPoint, MultiPolygon, Polygon
        if isinstance(shape, Point):
            return self._intersects_point(shape)
        elif isinstance(shape, MultiPoint):
            return self._intersects_multipoint(shape)
        elif isinstance(shape, Line):
            return self._intersects_line(shape)
        elif isinstance(shape, MultiLine):
            return self._intersects_line(shape)
        elif isinstance(shape, Polygon):
            return self._intersects_polygon(shape)
        elif isinstance(shape, MultiPolygon):
            return self._intersects_polygon(shape)
        else:
            raise ValueError("Unsupported intersection type %s" % type(shape).__name__)


class PointArray(GeometryFixedArray):
    _element_type = Point

    @property
    def _dtype_class(self):
        return PointDtype

    @classmethod
    def from_geopandas(cls, ga):
        """
        Build a spatialpandas MultiPointArray from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries of MultiPoint or
            Point shapes.

        Returns:
            MultiPointArray
        """
        return super().from_geopandas(ga)

    @property
    def length(self):
        return np.zeros(len(self), dtype=np.float64)

    @property
    def area(self):
        return np.zeros(len(self), dtype=np.float64)

    @property
    def x(self):
        res = np.full(len(self), np.nan)
        mask = ~self.isna()
        res[mask] = self.flat_values[0::2][mask]
        return res

    @property
    def y(self):
        res = np.full(len(self), np.nan)
        mask = ~self.isna()
        res[mask] = self.flat_values[1::2][mask]
        return res

    def intersects_bounds(self, bounds, inds=None):
        x0, y0, x1, y1 = bounds
        if x1 < x0:
            x0, x1 = x1, x0
        if y1 < y0:
            y0, y1 = y1, y0

        xs = self.x
        ys = self.y
        if inds is not None:
            xs = xs[inds]
            ys = ys[inds]

        outside = np.isnan(xs) | (xs < x0) | (xs > x1) | (ys < y0) | (ys > y1)
        return ~outside

    def _intersects_point(self, point, inds):
        flat = self.flat_values
        if inds is None:
            return (flat[0::2] == point.x) & (flat[1::2] == point.y)
        else:
            return (flat[inds * 2] == point.x) & (flat[inds * 2 + 1] == point.y)

    def _intersects_multipoint(self, multipoint, inds):
        flat_points = self.flat_values
        flat_multipoint = multipoint.flat_values
        if inds is None:
            inds = np.arange(len(self))
        return _perform_intersects_multipoint(flat_points, flat_multipoint, inds)

    def _intersects_line(self, line, inds):
        if inds is None:
            inds = np.arange(len(self))
        return _perform_intersects_line(
            self.flat_values, line.buffer_values,  line.buffer_inner_offsets, inds
        )

    def _intersects_polygon(self, polygon, inds):
        if inds is None:
            inds = np.arange(len(self))
        return _perform_intersects_polygon(
            self.flat_values, polygon.buffer_values,  polygon.buffer_inner_offsets, inds
        )

    def intersects(self, shape, inds=None):
        from . import Line, MultiLine, MultiPoint, MultiPolygon, Polygon
        if isinstance(shape, Point):
            return self._intersects_point(shape, inds)
        elif isinstance(shape, MultiPoint):
            return self._intersects_multipoint(shape, inds)
        elif isinstance(shape, Line):
            return self._intersects_line(shape, inds)
        elif isinstance(shape, MultiLine):
            return self._intersects_line(shape, inds)
        elif isinstance(shape, Polygon):
            return self._intersects_polygon(shape, inds)
        elif isinstance(shape, MultiPolygon):
            return self._intersects_polygon(shape, inds)
        else:
            raise ValueError("Unsupported intersection type %s" % type(shape).__name__)


@ngpjit
def _perform_intersects_multipoint(flat_points, flat_multipoint, inds):
    n = len(inds)
    multi_xs = flat_multipoint[0::2]
    multi_ys = flat_multipoint[1::2]
    result = np.zeros(n, dtype=np.bool_)
    for i, j in enumerate(inds):
        x = flat_points[2 * j]
        y = flat_points[2 * j + 1]
        result[i] = np.any((multi_xs == x) & (multi_ys == y))

    return result


@ngpjit
def _perform_intersects_line(flat_points, flat_lines, offsets, inds):
    n = len(inds)
    result = np.zeros(n, dtype=np.bool_)
    for i, j in enumerate(inds):
        x = flat_points[2 * j]
        y = flat_points[2 * j + 1]

        for k in range(len(offsets) - 1):
            flat_line = flat_lines[offsets[k]:offsets[k + 1]]
            line_xs = flat_line[0::2]
            line_ys = flat_line[1::2]
            bounds = (min(line_xs), min(line_ys), max(line_xs), max(line_ys))

            # Check bounding box
            if x < bounds[0] or y < bounds[1] or x > bounds[2] or y > bounds[3]:
                continue

            # Check line vertices
            intersects_vert = np.any((line_xs == x) & (line_ys == y))
            if intersects_vert:
                result[i] = True
                continue

            # Check whether point is on a line segment
            for m in range(len(line_xs) - 1):
                ax0 = line_xs[m]
                ay0 = line_ys[m]
                ax1 = line_xs[m + 1]
                ay1 = line_ys[m + 1]
                intersects_segment = segment_intersects_point(ax0, ay0, ax1, ay1, x, y)
                if intersects_segment:
                    result[i] = True
                    break

    return result


@ngpjit
def _perform_intersects_polygon(flat_points, flat_polygons, offsets, inds):
    n = len(inds)
    result = np.zeros(n, dtype=np.bool_)
    for i, j in enumerate(inds):
        x = flat_points[2 * j]
        y = flat_points[2 * j + 1]

        result[i] = point_intersects_polygon(
            x, y, flat_polygons, offsets
        )
    return result


def _points_array_non_empty(dtype):
    """
    Create an example length 2 array to register with Dask.
    See https://docs.dask.org/en/latest/dataframe-extend.html#extension-arrays
    """
    return PointArray([
        [1, 0],
        [1, 2]
    ], dtype=dtype)


if make_array_nonempty:
    make_array_nonempty.register(PointDtype)(_points_array_non_empty)



import numpy as np
import pandas as pd
import shapely.geometry as sg
from geopandas.array import from_shapely
from hypothesis import example, given
from hypothesis import strategies as st

from ..strategies import (
    hyp_settings,
    st_line_array,
    st_multipoint_array,
    st_multipolygon_array,
    st_point_array,
    st_polygon_array,
)
from spatialpandas import GeoSeries
from spatialpandas.geometry import (
    Line,
    MultiLine,
    MultiPoint,
    MultiPolygon,
    Point,
    PointArray,
    Polygon,
)


@given(st_point_array(), st_point_array(min_size=1, max_size=1))
@hyp_settings
def test_points_intersects_point(gp_points, gp_point):
    # Get scalar Point
    sg_point = gp_point[0]
    if len(gp_points) > 0:
        # Add gp_point to gp_points so we know something will intersect
        gp_points = pd.concat([pd.Series(gp_points), pd.Series(gp_point)]).array

    # Compute expected intersection
    expected = gp_points.intersects(sg_point)

    # Create spatialpandas PointArray
    point = Point.from_shapely(sg_point)
    points = PointArray.from_geopandas(gp_points)
    points_series = GeoSeries(points, index=np.arange(10, 10 + len(points)))

    # Test Point.intersects
    result = np.array([
        point_el.intersects(point) for point_el in points
    ])
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersect
    result = points.intersects(point)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(point, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test GeoSeries.intersects
    pd.testing.assert_series_equal(
        points_series.intersects(point),
        pd.Series(expected, index=points_series.index)
    )


@given(
    st_point_array(min_size=6),
    st.integers(min_value=0, max_value=5),
    st_point_array(min_size=1, max_size=1)
)
@hyp_settings
def test_points_intersects_point_offset(gp_points, offset, gp_point):
    # Get scalar Point
    sg_point = gp_point[0]
    if len(gp_points) > 0:
        # Add gp_point to gp_points so we know something will intersect
        gp_points = pd.concat([pd.Series(gp_points), pd.Series(gp_point)]).array

    # Compute expected intersection
    expected = gp_points.intersects(sg_point)[offset:]

    # Create spatialpandas PointArray
    point = Point.from_shapely(sg_point)
    points = PointArray.from_geopandas(gp_points)[offset:]

    # Test PointArray.intersect
    result = points.intersects(point)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(point, inds)
    np.testing.assert_equal(result, np.flipud(expected))


@given(st_point_array(), st_multipoint_array(min_size=1, max_size=1))
@hyp_settings
def test_points_intersects_multipoint(gp_points, gp_multipoint):
    # Get scalar Point
    sg_multipoint = gp_multipoint[0]
    if len(gp_points) > 0:
        # Add gp_point to gp_multipoints so we know something will intersect
        gp_points = from_shapely(list(gp_points) + [gp_multipoint[0].geoms[-1]])

    # Compute expected intersection
    expected = gp_points.intersects(sg_multipoint)

    # Create spatialpandas PointArray
    multipoint = MultiPoint.from_shapely(sg_multipoint)
    points = PointArray.from_geopandas(gp_points)
    points_series = GeoSeries(points, index=np.arange(10, 10 + len(points)))

    # Test Point.intersects
    result = np.array([
        point_el.intersects(multipoint) for point_el in points
    ])
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersect
    result = points.intersects(multipoint)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(multipoint, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test GeoSeries.intersects
    pd.testing.assert_series_equal(
        points_series.intersects(multipoint),
        pd.Series(expected, index=points_series.index)
    )


@given(st_point_array(), st_line_array(min_size=1, max_size=1))
@example(
    from_shapely([
        sg.Point([0.25, 0.25]),  # on line
        sg.Point([1, 1]),  # on vertex
        sg.Point([1.01, 1.01])  # on ray, just past vertex
    ]),
    from_shapely([sg.LineString([(0, 0), (1, 1), (2, 0)])]),
)
@hyp_settings
def test_points_intersects_line(gp_points, gp_line):
    # Get scalar Line
    sg_line = gp_line[0]

    # Compute expected intersection
    expected = gp_points.intersects(sg_line)

    # Create spatialpandas objects
    line = Line.from_shapely(sg_line)
    points = PointArray.from_geopandas(gp_points)
    points_series = GeoSeries(points, index=np.arange(10, 10 + len(points)))

    # Test Point.intersects
    result = np.array([
        point_el.intersects(line) for point_el in points
    ])
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersect
    result = points.intersects(line)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(line, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test GeoSeries.intersects
    pd.testing.assert_series_equal(
        points_series.intersects(line),
        pd.Series(expected, index=points_series.index)
    )


@given(st_point_array(), st_line_array(min_size=1, max_size=1))
@example(
    from_shapely([
        sg.Point([0.25, 0.25]),  # on line
        sg.Point([1, 1]),  # on vertex
        sg.Point([1.01, 1.01])  # on ray, just past vertex
    ]),
    from_shapely([sg.MultiLineString([
        [(1, 0.5), (2, 0)],
        [(0, 0), (1, 1)],
    ])])
)
@hyp_settings
def test_points_intersects_multiline(gp_points, gp_multiline):
    # Get scalar MultiLine
    sg_multiline = gp_multiline[0]

    # Compute expected intersection
    expected = gp_points.intersects(sg_multiline)

    # Create spatialpandas objects
    multiline = MultiLine.from_shapely(sg_multiline)
    points = PointArray.from_geopandas(gp_points)
    points_series = GeoSeries(points, index=np.arange(10, 10 + len(points)))

    # Test Point.intersects
    result = np.array([
        point_el.intersects(multiline) for point_el in points
    ])
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersect
    result = points.intersects(multiline)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(multiline, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test GeoSeries.intersects
    pd.testing.assert_series_equal(
        points_series.intersects(multiline),
        pd.Series(expected, index=points_series.index)
    )


@given(st_point_array(), st_polygon_array(min_size=1, max_size=1))
@hyp_settings
def test_points_intersects_polygon(gp_points, gp_polygon):
    # Get scalar Polygon
    sg_polygon = gp_polygon[0]

    # Compute expected intersection
    expected = gp_points.intersects(sg_polygon)

    # Create spatialpandas objects
    polygon = Polygon.from_shapely(sg_polygon)
    points = PointArray.from_geopandas(gp_points)
    points_series = GeoSeries(points, index=np.arange(10, 10 + len(points)))

    # Test Point.intersects
    result = np.array([
        point_el.intersects(polygon) for point_el in points
    ])
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersect
    result = points.intersects(polygon)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(polygon, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test GeoSeries.intersects
    pd.testing.assert_series_equal(
        points_series.intersects(polygon),
        pd.Series(expected, index=points_series.index)
    )


@given(st_point_array(), st_multipolygon_array(min_size=1, max_size=1))
@hyp_settings
def test_points_intersects_multipolygon(gp_points, gp_multipolygon):
    # Get scalar MultiPolygon
    sg_multipolygon = gp_multipolygon[0]

    # Compute expected intersection
    expected = gp_points.intersects(sg_multipolygon)

    # Create spatialpandas objects
    multipolygon = MultiPolygon.from_shapely(sg_multipolygon)
    points = PointArray.from_geopandas(gp_points)
    points_series = GeoSeries(points, index=np.arange(10, 10 + len(points)))

    # Test Point.intersects
    result = np.array([
        point_el.intersects(multipolygon) for point_el in points
    ])
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersect
    result = points.intersects(multipolygon)
    np.testing.assert_equal(result, expected)

    # Test PointArray.intersects with inds
    inds = np.flipud(np.arange(0, len(points)))
    result = points.intersects(multipolygon, inds)
    np.testing.assert_equal(result, np.flipud(expected))

    # Test GeoSeries.intersects
    pd.testing.assert_series_equal(
        points_series.intersects(multipolygon),
        pd.Series(expected, index=points_series.index)
    )



import warnings

import numpy as np
from numba import jit

ngjit = jit(nopython=True, nogil=True)
ngpjit = jit(nopython=True, nogil=True, parallel=True)


@ngjit
def _data2coord(vals, val_range, n):
    """
    Convert an array of values from continuous data coordinates to discrete
    integer coordinates

    Args:
        vals: Array of continuous data coordinate to be converted to discrete
            integer coordinates
        val_range: Tuple of start (val_range[0]) and stop (val_range[1]) range in
            continuous data coordinates
        n: The integer number of discrete distance coordinates

    Returns:
        unsigned integer array of discrete coordinates
    """
    x_width = val_range[1] - val_range[0]
    res = ((vals - val_range[0]) * (n / x_width)).astype(np.int64)

    # clip
    res[res < 0] = 0
    res[res > n - 1] = n - 1
    return res


def _asarray_maybe_ragged(input):
    """Convert input into a single numpy array, even if it is a ragged array.

    Prior to numpy 1.24 just np.asarray(input) suffices as it emits a
    np.VisibleDeprecationWarning if the input is ragged, i.e. can't be
    converted to a single numpy array, but still completes the conversion by
    creating a numpy array of multiple subarrays. From 1.24 onwards attempting
    to do this raises a ValueError instead. This function therefore tries the
    simple conversion and if this fails uses the ragged-supporting conversion
    of np.asarray(input, type=object).

    To demonstrate that this works with numpy < 1.24 it converts
    VisibleDeprecationWarnings into errors so that they are handled the same
    as for numpy >= 1.24.

    Args:
        input: ArrayLike | list[ArrayLike | None]

    Returns:
        NumPy array.
    """
    with warnings.catch_warnings():
        warnings.simplefilter('error', np.VisibleDeprecationWarning)
        try:
            array = np.asarray(input)
        except (ValueError, np.VisibleDeprecationWarning):
            array = np.asarray(input, dtype=object)

    return array



import numpy as np
from dask.dataframe.extensions import make_array_nonempty
from pandas.core.dtypes.dtypes import register_extension_dtype

from ..geometry._algorithms.intersection import (lines_intersect_bounds,
                                                 multilines_intersect_bounds)
from ..geometry._algorithms.measures import compute_line_length
from ..geometry.base import GeometryDtype
from ..geometry.baselist import (GeometryList, GeometryListArray,
                                 _geometry_map_nested2)


@register_extension_dtype
class MultiLineDtype(GeometryDtype):
    _geometry_name = 'multiline'
    @classmethod
    def construct_array_type(cls, *args):
        if len(args) > 0:
            raise NotImplementedError("construct_array_type does not support arguments")
        return MultiLineArray


class MultiLine(GeometryList):
    _nesting_levels = 1

    @classmethod
    def construct_array_type(cls):
        return MultiLineArray

    @classmethod
    def _shapely_to_coordinates(cls, shape):
        import shapely.geometry as sg
        if isinstance(shape, sg.MultiLineString):
            shape = shape.geoms
            line_parts = []
            if len(shape) == 0:
                # Add single empty line so we have the right number of nested levels
                line_parts.append([])
            else:
                for line in shape:
                    line_parts.append(np.asarray(line.coords).ravel())
            return line_parts
        elif isinstance(shape, (sg.LineString, sg.LinearRing)):
            return [np.asarray(shape.coords).ravel()]
        else:
            raise ValueError("""
Received invalid value of type {typ}. Must be an instance of MultiLineString
""".format(typ=type(shape).__name__))

    def to_shapely(self):
        """
        Convert to shapely shape

        Returns:
            shapely MultiLineString shape
        """
        import shapely.geometry as sg

        def to_numpy_2d(v):
            a = np.array(v.as_py(), dtype=self.numpy_dtype)
            return a.reshape(len(v) // 2, 2)

        line_arrays = [to_numpy_2d(line_coords) for line_coords in self.data]
        lines = [sg.LineString(line_array) for line_array in line_arrays]
        return sg.MultiLineString(lines=lines)

    @classmethod
    def from_shapely(cls, shape):
        """
        Build a spatialpandas MultiLine object from a shapely shape

        Args:
            shape: A shapely MultiLineString, LineString, or LinearRing shape

        Returns:
            spatialpandas MultiLine
        """
        return super().from_shapely(shape)

    @property
    def length(self):
        return compute_line_length(self.buffer_values, self.buffer_inner_offsets)

    @property
    def area(self):
        return 0.0

    def intersects_bounds(self, bounds):
        x0, y0, x1, y1 = bounds
        offsets = self.buffer_outer_offsets
        start_offsets = offsets[:-1]
        stop_offstes = offsets[1:]
        result = np.zeros(len(start_offsets), dtype=np.bool_)
        lines_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1),
            self.buffer_values, start_offsets, stop_offstes, result
        )
        return result.any()


class MultiLineArray(GeometryListArray):
    _element_type = MultiLine
    _nesting_levels = 2

    @property
    def _dtype_class(self):
        return MultiLineDtype

    @classmethod
    def from_geopandas(cls, ga):
        """
        Build a spatialpandas MultiLineArray from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries of MultiLineString,
            LineString, or LinearRing shapes.

        Returns:
            MultiLineArray
        """
        return super().from_geopandas(ga)

    @property
    def length(self):
        result = np.full(len(self), np.nan, dtype=np.float64)
        _geometry_map_nested2(
            compute_line_length,
            result,
            self.buffer_values,
            self.buffer_offsets,
            self.isna(),
        )
        return result

    @property
    def area(self):
        return np.zeros(len(self), dtype=np.float64)

    def intersects_bounds(self, bounds, inds=None):
        x0, y0, x1, y1 = bounds
        offsets0, offsets1 = self.buffer_offsets
        start_offsets0 = offsets0[:-1]
        stop_offsets0 = offsets0[1:]
        if inds is not None:
            start_offsets0 = start_offsets0[inds]
            stop_offsets0 = stop_offsets0[inds]

        result = np.zeros(len(start_offsets0), dtype=np.bool_)
        multilines_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1),
            self.buffer_values, start_offsets0, stop_offsets0, offsets1, result
        )
        return result


def _multi_line_array_non_empty(dtype):
    """
    Create an example length 2 array to register with Dask.
    See https://docs.dask.org/en/latest/dataframe-extend.html#extension-arrays
    """
    return MultiLineArray([
        [[1, 0, 1, 1], [1, 2, 0, 0]],
        [[3, 3, 4, 4]]
    ], dtype=dtype)


if make_array_nonempty:
    make_array_nonempty.register(MultiLineDtype)(_multi_line_array_non_empty)



import numpy as np
from dask.dataframe.extensions import make_array_nonempty
from pandas.core.dtypes.dtypes import register_extension_dtype

from ..geometry._algorithms.intersection import lines_intersect_bounds
from ..geometry._algorithms.measures import compute_line_length
from ..geometry.base import GeometryDtype
from ..geometry.baselist import (GeometryList, GeometryListArray,
                                 _geometry_map_nested1)


@register_extension_dtype
class LineDtype(GeometryDtype):
    _geometry_name = 'line'

    @classmethod
    def construct_array_type(cls, *args):
        return LineArray


class Line(GeometryList):
    _nesting_levels = 0

    @classmethod
    def construct_array_type(cls):
        return LineArray

    @classmethod
    def _shapely_to_coordinates(cls, shape):
        import shapely.geometry as sg
        if isinstance(shape, (sg.LineString, sg.LinearRing)):
            # Single line
            return np.asarray(shape.coords).ravel()
        else:
            raise ValueError("""
Received invalid value of type {typ}. Must be an instance of LineString
or LinearRing""".format(typ=type(shape).__name__))

    def to_shapely(self):
        """
        Convert to shapely shape

        Returns:
            shapely LineString shape
        """
        import shapely.geometry as sg
        line_coords = np.array(self.data.as_py(), dtype=self.numpy_dtype)
        return sg.LineString(line_coords.reshape(len(line_coords) // 2, 2))

    @classmethod
    def from_shapely(cls, shape):
        """
        Build a spatialpandas Line object from a shapely shape

        Args:
            shape: A shapely LineString or LinearRing shape

        Returns:
            spatialpandas Line
        """
        return super().from_shapely(shape)

    @property
    def length(self):
        return compute_line_length(self.buffer_values, self.buffer_inner_offsets)

    @property
    def area(self):
        return 0.0

    def intersects_bounds(self, bounds):
        x0, y0, x1, y1 = bounds
        result = np.zeros(1, dtype=np.bool_)
        offsets = self.buffer_outer_offsets
        lines_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1),
            self.buffer_values, offsets[:-1], offsets[1:], result
        )
        return result[0]


class LineArray(GeometryListArray):
    _element_type = Line
    _nesting_levels = 1

    @property
    def _dtype_class(self):
        return LineDtype

    @classmethod
    def from_geopandas(cls, ga):
        """
        Build a spatialpandas LineArray from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries of `LineString` or
            `LinearRing`shapes.

        Returns:
            LineArray
        """
        return super().from_geopandas(ga)

    @property
    def length(self):
        result = np.full(len(self), np.nan, dtype=np.float64)
        _geometry_map_nested1(
            compute_line_length,
            result,
            self.buffer_values,
            self.buffer_offsets,
            self.isna(),
        )
        return result

    @property
    def area(self):
        return np.zeros(len(self), dtype=np.float64)

    def intersects_bounds(self, bounds, inds=None):
        x0, y0, x1, y1 = bounds
        offsets = self.buffer_outer_offsets
        start_offsets0 = offsets[:-1]
        stop_offsets0 = offsets[1:]
        if inds is not None:
            start_offsets0 = start_offsets0[inds]
            stop_offsets0 = stop_offsets0[inds]

        result = np.zeros(len(stop_offsets0), dtype=np.bool_)
        lines_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1),
            self.buffer_values, start_offsets0, stop_offsets0, result
        )
        return result


def _line_array_non_empty(dtype):
    """
    Create an example length 2 array to register with Dask.
    See https://docs.dask.org/en/latest/dataframe-extend.html#extension-arrays
    """
    return LineArray([
        [1, 0, 1, 1],
        [1, 2, 0, 0]
    ], dtype=dtype)


if make_array_nonempty:
    make_array_nonempty.register(LineDtype)(_line_array_non_empty)



import pandas as pd

from .geometry import GeometryDtype, Geometry


class _MaybeGeoSeries(pd.Series):
    def __new__(cls, data, *args, **kwargs):
        if isinstance(getattr(data, 'dtype', None), GeometryDtype):
            series_cls = GeoSeries
        else:
            series_cls = pd.Series
        return series_cls(data, *args, **kwargs)


class GeoSeries(pd.Series):
    def __init__(self, data, index=None, name=None, dtype=None, **kwargs):
        from .geometry.base import to_geometry_array

        # Handle scalar geometry with index
        if isinstance(data, Geometry):
            n = len(index) if index is not None else 1
            data = [data] * n

        if index is None and hasattr(data, 'dtype'):
            # Try to get input index from Series-like object
            index = getattr(data, 'index', None)
        if name is None:
            name = getattr(data, 'name', None)

        # Normalize dtype from string
        if dtype is not None:
            dtype = pd.array([], dtype=dtype).dtype

        data = to_geometry_array(data, dtype)
        super().__init__(data, index=index, name=name, **kwargs)

    @property
    def _constructor(self):
        return _MaybeGeoSeries

    @property
    def _constructor_expanddim(self):
        from .geodataframe import GeoDataFrame
        return GeoDataFrame

    @property
    def bounds(self):
        return pd.DataFrame(
            self.array.bounds, columns=['x0', 'y0', 'x1', 'y1'], index=self.index
        )

    @property
    def total_bounds(self):
        return self.array.total_bounds

    @property
    def area(self):
        return pd.Series(self.array.area, index=self.index)

    @property
    def length(self):
        return pd.Series(self.array.length, index=self.index)

    def hilbert_distance(self, total_bounds=None, p=15):
        return pd.Series(
            self.array.hilbert_distance(total_bounds=total_bounds, p=p),
            index=self.index
        )

    @property
    def sindex(self):
        return self.array.sindex

    @property
    def cx(self):
        from .geometry.base import _CoordinateIndexer
        return _CoordinateIndexer(self.array, parent=self)

    def build_sindex(self, **kwargs):
        self.array.build_sindex(**kwargs)
        return self

    def intersects_bounds(self, bounds):
        return pd.Series(
            self.array.intersects_bounds(bounds), index=self.index
        )

    def intersects(self, shape):
        return pd.Series(
            self.array.intersects(shape), index=self.index
        )

    def to_geopandas(self):
        from geopandas import GeoSeries
        return GeoSeries(self.array.to_geopandas(), index=self.index)



import copy
import json
import os
import uuid
from inspect import signature

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from retrying import retry

import dask
import dask.dataframe as dd
from dask import delayed
from dask.dataframe.core import get_parallel_type
from dask.dataframe.partitionquantiles import partition_quantiles
from dask.dataframe.extensions import make_array_nonempty
try:
    from dask.dataframe.dispatch import make_meta_dispatch
    from dask.dataframe.backends import meta_nonempty
except ImportError:
    from dask.dataframe.utils import make_meta as make_meta_dispatch, meta_nonempty

from .geodataframe import GeoDataFrame
from .geometry.base import GeometryDtype, _BaseCoordinateIndexer
from .geoseries import GeoSeries
from .spatialindex import HilbertRtree


class DaskGeoSeries(dd.Series):
    def __init__(self, dsk, name, meta, divisions, *args, **kwargs):
        super().__init__(dsk, name, meta, divisions)

        # Init backing properties
        self._partition_bounds = None
        self._partition_sindex = None

    @property
    def bounds(self):
        return self.map_partitions(lambda s: s.bounds)

    @property
    def total_bounds(self):
        partition_bounds = self.partition_bounds
        return (
            np.nanmin(partition_bounds['x0']),
            np.nanmin(partition_bounds['y0']),
            np.nanmax(partition_bounds['x1']),
            np.nanmax(partition_bounds['y1']),
        )

    @property
    def partition_bounds(self):
        if self._partition_bounds is None:
            self._partition_bounds = self.map_partitions(
                lambda s: pd.DataFrame(
                    [s.total_bounds], columns=['x0', 'y0', 'x1', 'y1']
                )
            ).compute().reset_index(drop=True)
            self._partition_bounds.index.name = 'partition'
        return self._partition_bounds

    @property
    def area(self):
        return self.map_partitions(lambda s: s.area)

    @property
    def length(self):
        return self.map_partitions(lambda s: s.length)

    @property
    def partition_sindex(self):
        if self._partition_sindex is None:
            self._partition_sindex = HilbertRtree(self.partition_bounds.values)
        return self._partition_sindex

    @property
    def cx(self):
        return _DaskCoordinateIndexer(self, self.partition_sindex)

    @property
    def cx_partitions(self):
        return _DaskPartitionCoordinateIndexer(self, self.partition_sindex)

    def build_sindex(self, **kwargs):
        def build_sindex(series, **kwargs):
            series.build_sindex(**kwargs)
            return series
        return self.map_partitions(build_sindex, **kwargs, meta=self._meta)

    def intersects_bounds(self, bounds):
        return self.map_partitions(lambda s: s.intersects_bounds(bounds))

    # Override some standard Dask Series methods to propagate extra properties
    def _propagate_props_to_series(self, new_series):
        new_series._partition_bounds = self._partition_bounds
        new_series._partition_sindex = self._partition_sindex
        return new_series

    def persist(self, **kwargs):
        return self._propagate_props_to_series(
            super().persist(**kwargs)
        )


@make_meta_dispatch.register(GeoSeries)
def make_meta_series(s, index=None):
    result = s.head(0)
    if index is not None:
        result = result.reindex(index[:0])
    return result


@meta_nonempty.register(GeoSeries)
def meta_nonempty_series(s, index=None):
    return GeoSeries(make_array_nonempty(s.dtype), index=index)


@get_parallel_type.register(GeoSeries)
def get_parallel_type_dataframe(df):
    return DaskGeoSeries


class DaskGeoDataFrame(dd.DataFrame):
    def __init__(self, dsk, name, meta, divisions):
        super().__init__(dsk, name, meta, divisions)
        self._partition_sindex = {}
        self._partition_bounds = {}

    def to_parquet(self, fname, compression="snappy", filesystem=None, **kwargs):
        from .io import to_parquet_dask
        to_parquet_dask(
            self, fname, compression=compression, filesystem=filesystem, **kwargs
        )

    @property
    def geometry(self):
        # Use self._meta.geometry.name rather than self._meta._geometry so that an
        # informative error message is raised if there is no valid geometry column
        return self[self._meta.geometry.name]

    def set_geometry(self, geometry):
        if geometry != self._meta._geometry:
            return self.map_partitions(lambda df: df.set_geometry(geometry))
        else:
            return self

    @property
    def partition_sindex(self):
        geometry_name = self._meta.geometry.name
        if geometry_name not in self._partition_sindex:

            # Apply partition_bounds to geometry Series before creating the spatial
            # index. This removes the need to scan partitions to compute bounds
            geometry = self.geometry
            if geometry_name in self._partition_bounds:
                geometry._partition_bounds = self._partition_bounds[geometry_name]

            self._partition_sindex[geometry.name] = geometry.partition_sindex
            self._partition_bounds[geometry_name] = geometry.partition_bounds
        return self._partition_sindex[geometry_name]

    @property
    def cx(self):
        return _DaskCoordinateIndexer(self, self.partition_sindex)

    @property
    def cx_partitions(self):
        return _DaskPartitionCoordinateIndexer(self, self.partition_sindex)

    def pack_partitions(self, npartitions=None, p=15, shuffle='tasks'):
        """
        Repartition and reorder dataframe spatially along a Hilbert space filling curve

        Args:
            npartitions: Number of output partitions. Defaults to the larger of 8 and
                the length of the dataframe divided by 2**23.
            p: Hilbert curve p parameter
            shuffle: Dask shuffle method, either "disk" or "tasks"

        Returns:
            Spatially partitioned DaskGeoDataFrame
        """
        # Compute number of output partitions
        npartitions = self._compute_packing_npartitions(npartitions)

        # Add hilbert_distance column
        ddf = self._with_hilbert_distance_column(p)

        # Set index to distance. This will trigger an expensive shuffle
        # sort operation
        ddf = ddf.set_index('hilbert_distance', npartitions=npartitions, shuffle=shuffle)

        if ddf.npartitions != npartitions:
            # set_index doesn't change the number of partitions if the partitions
            # happen to be already sorted
            ddf = ddf.repartition(npartitions=npartitions)

        return ddf

    def pack_partitions_to_parquet(
        self,
        path,
        filesystem=None,
        npartitions=None,
        p=15,
        compression="snappy",
        tempdir_format=None,
        _retry_args=None,
        storage_options=None,
        engine_kwargs=None,
        overwrite=False,
    ):
        """
        Repartition and reorder dataframe spatially along a Hilbert space filling curve
        and write to parquet dataset at the provided path.

        This is equivalent to ddf.pack_partitions(...).to_parquet(...) but with lower
        memory and disk usage requirements.

        Args:
            path: Output parquet dataset path
            filesystem: Optional fsspec filesystem. If not provided, filesystem type
                is inferred from path
            npartitions: Number of output partitions. Defaults to the larger of 8 and
                the length of the dataframe divided by 2**23.
            p: Hilbert curve p parameter
            compression: Compression algorithm for parquet file
            tempdir_format: format string used to generate the filesystem path where
                temporary files should be stored for each output partition.  String
                must contain a '{partition}' replacement field which will be formatted
                using the output partition number as an integer. The string may
                optionally contain a '{uuid}' replacement field which will be formatted
                using a randomly generated UUID string. If None (the default),
                temporary files are stored inside the output path.

                These directories are deleted as soon as possible during the execution
                of the function.
            storage_options: Key/value pairs to be passed on to the file-system backend, if any.
            engine_kwargs: pyarrow.parquet engine-related keyword arguments.
        Returns:
            DaskGeoDataFrame backed by newly written parquet dataset
        """
        from .io import read_parquet, read_parquet_dask
        from .io.utils import validate_coerce_filesystem

        engine_kwargs = engine_kwargs or {}

        # Get fsspec filesystem object
        filesystem = validate_coerce_filesystem(path, filesystem, storage_options)

        # Decorator for operations that should be retried
        if _retry_args is None:
            _retry_args = dict(
                wait_exponential_multiplier=100,
                wait_exponential_max=120000,
                stop_max_attempt_number=24,
            )
        retryit = retry(**_retry_args)

        @retryit
        def rm_retry(file_path):
            filesystem.invalidate_cache()
            if filesystem.exists(file_path):
                filesystem.rm(file_path, recursive=True)
                if filesystem.exists(file_path):
                    # Make sure we keep retrying until file does not exist
                    raise ValueError("Deletion of {path} not yet complete".format(
                        path=file_path
                    ))

        @retryit
        def mkdirs_retry(dir_path):
            filesystem.makedirs(dir_path, exist_ok=True)

        # For filesystems that provide a "refresh" argument, set it to True
        if 'refresh' in signature(filesystem.ls).parameters:
            ls_kwargs = {'refresh': True}
        else:
            ls_kwargs = {}

        @retryit
        def ls_retry(dir_path):
            filesystem.invalidate_cache()
            return filesystem.ls(dir_path, **ls_kwargs)

        @retryit
        def move_retry(p1, p2):
            if filesystem.exists(p1):
                filesystem.move(p1, p2)

        # Compute tempdir_format string
        dataset_uuid = str(uuid.uuid4())
        if tempdir_format is None:
            tempdir_format = os.path.join(path, "part.{partition}.parquet")
        elif not isinstance(tempdir_format, str) or "{partition" not in tempdir_format:
            raise ValueError(
                "tempdir_format must be a string containing a {{partition}} "
                "replacement field\n"
                "    Received: {tempdir_format}".format(
                    tempdir_format=repr(tempdir_format)
                )
            )

        # Compute number of output partitions
        npartitions = self._compute_packing_npartitions(npartitions)
        out_partitions = list(range(npartitions))

        # Add hilbert_distance column
        ddf = self._with_hilbert_distance_column(p)

        # Compute output hilbert_distance divisions
        quantiles = partition_quantiles(
            ddf.hilbert_distance, npartitions
        ).compute().values

        # Add _partition column containing output partition number of each row
        ddf = ddf.map_partitions(
            lambda df: df.assign(
                _partition=np.digitize(df.hilbert_distance, quantiles[1:], right=True))
        )

        # Compute part paths
        parts_tmp_paths = [
            tempdir_format.format(partition=out_partition, uuid=dataset_uuid)
            for out_partition in out_partitions
        ]
        part_output_paths = [
            os.path.join(path, "part.%d.parquet" % out_partition)
            for out_partition in out_partitions
        ]

        # Initialize output partition directory structure
        filesystem.invalidate_cache()
        if overwrite:
            rm_retry(path)

        for out_partition in out_partitions:
            part_dir = os.path.join(path, "part.%d.parquet" % out_partition)
            mkdirs_retry(part_dir)
            tmp_part_dir = tempdir_format.format(partition=out_partition, uuid=dataset_uuid)
            mkdirs_retry(tmp_part_dir)

        # Shuffle and write a parquet dataset for each output partition
        @retryit
        def write_partition(df_part, part_path):
            with filesystem.open(part_path, "wb") as f:
                df_part.to_parquet(
                    f,
                    compression=compression,
                    index=True,
                    **engine_kwargs,
                )


        def process_partition(df, i):
            subpart_paths = {}
            for out_partition, df_part in df.groupby('_partition'):
                part_path = os.path.join(
                    tempdir_format.format(partition=out_partition, uuid=dataset_uuid),
                    'part.%d.parquet' % i,
                )
                df_part = (
                    df_part
                    .drop('_partition', axis=1)
                    .set_index('hilbert_distance', drop=True)
                )
                write_partition(df_part, part_path)
                subpart_paths[out_partition] = part_path

            return subpart_paths

        part_path_infos = dask.compute(*[
            dask.delayed(process_partition, pure=False)(df, i)
            for i, df in enumerate(ddf.to_delayed())
        ])

        # Build dict from part number to list of subpart paths
        part_num_to_subparts = {}
        for part_path_info in part_path_infos:
            for part_num, subpath in part_path_info.items():
                subpaths = part_num_to_subparts.get(part_num, [])
                subpaths.append(subpath)
                part_num_to_subparts[part_num] = subpaths

        # Concat parquet dataset per partition into parquet file per partition
        @retryit
        def write_concatted_part(part_df, part_output_path, md_list):
            with filesystem.open(part_output_path, 'wb') as f:
                pq.write_table(
                    pa.Table.from_pandas(part_df),
                    f, compression=compression, metadata_collector=md_list
                )

        @retryit
        def read_parquet_retry(parts_tmp_path, subpart_paths, part_output_path):
            if filesystem.isfile(part_output_path) and not filesystem.isdir(parts_tmp_path):
                # Handle rare case where the task was resubmitted and the work has
                # already been done.  This shouldn't happen with pure=False, but it
                # seems like it does very rarely.
                return read_parquet(
                    part_output_path,
                    filesystem=filesystem,
                    storage_options=storage_options,
                    **engine_kwargs,
                )

            ls_res = sorted(filesystem.ls(parts_tmp_path, **ls_kwargs))
            subpart_paths_stripped = sorted([filesystem._strip_protocol(_) for _ in subpart_paths])

            if subpart_paths_stripped != ls_res:
                missing = set(subpart_paths) - set(ls_res)
                extras = set(ls_res) - set(subpart_paths)
                raise ValueError(
                    "Filesystem not yet consistent\n"
                    "  Expected len: {expected}\n"
                    "  Found len: {received}\n"
                    "  Missing: {missing}\n"
                    "  Extras: {extras}".format(
                        expected=len(subpart_paths),
                        received=len(ls_res),
                        missing=list(missing),
                        extras=list(extras)
                    )
                )
            return read_parquet(
                parts_tmp_path,
                filesystem=filesystem,
                storage_options=storage_options,
                **engine_kwargs,
            )

        def concat_parts(parts_tmp_path, subpart_paths, part_output_path):
            filesystem.invalidate_cache()

            # Load directory of parquet parts for this partition into a
            # single GeoDataFrame
            if not subpart_paths:
                # Empty partition
                rm_retry(parts_tmp_path)
                return None
            else:
                part_df = read_parquet_retry(parts_tmp_path, subpart_paths, part_output_path)

            # Compute total_bounds for all geometry columns in part_df
            total_bounds = {}
            for series_name in part_df.columns:
                series = part_df[series_name]
                if isinstance(series.dtype, GeometryDtype):
                    total_bounds[series_name] = series.total_bounds

            # Delete directory of parquet parts for partition
            rm_retry(parts_tmp_path)
            rm_retry(part_output_path)

            # Sort by part_df by hilbert_distance index
            part_df.sort_index(inplace=True)

            # Write part_df as a single parquet file, collecting metadata for later use
            # constructing the full dataset _metadata file.
            md_list = []
            filesystem.invalidate_cache()
            write_concatted_part(part_df, part_output_path, md_list)

            # Return metadata and total_bounds for part
            return {"meta": md_list[0], "total_bounds": total_bounds}

        write_info = dask.compute(*[
            dask.delayed(concat_parts, pure=False)(
                parts_tmp_paths[out_partition],
                part_num_to_subparts.get(out_partition, []),
                part_output_paths[out_partition]
            )
            for out_partition in out_partitions
        ])

        # Handle empty partitions.
        input_paths, write_info = zip(*[
            (pp, wi) for (pp, wi) in zip(part_output_paths, write_info) if wi is not None
        ])
        output_paths = part_output_paths[:len(input_paths)]
        for p1, p2 in zip(input_paths, output_paths):
            if p1 != p2:
                move_retry(p1, p2)

        # Write _metadata
        meta = write_info[0]['meta']
        for i in range(1, len(write_info)):
            meta.append_row_groups(write_info[i]["meta"])

        @retryit
        def write_metadata_file():
            with filesystem.open(os.path.join(path, "_metadata"), 'wb') as f:
                meta.write_metadata_file(f)
        write_metadata_file()

        # Collect total_bounds per partition for all geometry columns
        all_bounds = {}
        for info in write_info:
            for col, bounds in info.get('total_bounds', {}).items():
                bounds_list = all_bounds.get(col, [])
                bounds_list.append(
                    pd.Series(bounds, index=['x0', 'y0', 'x1', 'y1'])
                )
                all_bounds[col] = bounds_list

        # Build spatial metadata for parquet dataset
        partition_bounds = {}
        for col, bounds in all_bounds.items():
            partition_bounds[col] = pd.DataFrame(all_bounds[col]).to_dict()

        spatial_metadata = {'partition_bounds': partition_bounds}
        b_spatial_metadata = json.dumps(spatial_metadata).encode('utf')

        # Write _common_metadata
        @retryit
        def write_commonmetadata_file():
            with filesystem.open(os.path.join(path, "part.0.parquet")) as f:
                pf = pq.ParquetFile(f)

            all_metadata = copy.copy(pf.metadata.metadata)
            all_metadata[b'spatialpandas'] = b_spatial_metadata

            new_schema = pf.schema.to_arrow_schema().with_metadata(all_metadata)
            with filesystem.open(os.path.join(path, "_common_metadata"), 'wb') as f:
                pq.write_metadata(new_schema, f)
        write_commonmetadata_file()

        return read_parquet_dask(
            path,
            filesystem=filesystem,
            storage_options=storage_options,
            engine_kwargs=engine_kwargs,
        )

    def _compute_packing_npartitions(self, npartitions):
        if npartitions is None:
            # Make partitions of ~8 million rows with a minimum of 8
            # partitions
            nrows = len(self)
            npartitions = max(nrows // 2 ** 23, 8)
        return npartitions

    def _with_hilbert_distance_column(self, p):
        # Get geometry column
        geometry = self.geometry
        # Compute distance of points along the Hilbert-curve
        total_bounds = geometry.total_bounds
        ddf = self.assign(hilbert_distance=geometry.map_partitions(
            lambda s: s.hilbert_distance(total_bounds=total_bounds, p=p))
        )
        return ddf

    # Override some standard Dask Dataframe methods to propagate extra properties
    def _propagate_props_to_dataframe(self, new_frame):
        new_frame._partition_sindex = self._partition_sindex
        new_frame._partition_bounds = self._partition_bounds
        return new_frame

    def _propagate_props_to_series(self, new_series):
        if new_series.name in self._partition_bounds:
            new_series._partition_bounds = self._partition_bounds[new_series.name]
        if new_series.name in self._partition_sindex:
            new_series._partition_sindex = self._partition_sindex[new_series.name]
        return new_series

    def build_sindex(self, **kwargs):
        def build_sindex(df, **kwargs):
            df.build_sindex(**kwargs)
            return df
        return self.map_partitions(build_sindex, **kwargs, meta=self._meta)

    def persist(self, **kwargs):
        return self._propagate_props_to_dataframe(
            super().persist(**kwargs)
        )

    def __getitem__(self, key):
        result = super().__getitem__(key)
        if np.isscalar(key) or isinstance(key, (tuple, str)):
            # New series of single column, partition props apply if we have them
            self._propagate_props_to_series(result)
        elif isinstance(key, (np.ndarray, list)):
            # New dataframe with same length and same partitions as self so partition
            # properties still apply
            self._propagate_props_to_dataframe(result)
        return result


@make_meta_dispatch.register(GeoDataFrame)
def make_meta_dataframe(df, index=None):
    result = df.head(0)
    if index is not None:
        result = result.reindex(index[:0])
    return result


@meta_nonempty.register(GeoDataFrame)
def meta_nonempty_dataframe(df, index=None):
    return GeoDataFrame(meta_nonempty(pd.DataFrame(df.head(0))))


@get_parallel_type.register(GeoDataFrame)
def get_parallel_type_series(s):
    return DaskGeoDataFrame


class _DaskCoordinateIndexer(_BaseCoordinateIndexer):
    def __init__(self, obj, sindex):
        super().__init__(sindex)
        self._obj = obj

    def _perform_get_item(self, covers_inds, overlaps_inds, x0, x1, y0, y1):
        covers_inds = set(covers_inds)
        overlaps_inds = set(overlaps_inds)
        all_partition_inds = sorted(covers_inds.union(overlaps_inds))
        if len(all_partition_inds) == 0:
            # No partitions intersect with query region, return empty result
            return dd.from_pandas(self._obj._meta, npartitions=1)

        @delayed
        def cx_fn(df):
            return df.cx[x0:x1, y0:y1]

        ddf = self._obj.partitions[all_partition_inds]
        delayed_dfs = []
        for partition_ind, delayed_df in zip(all_partition_inds, ddf.to_delayed()):
            if partition_ind in overlaps_inds:
                delayed_dfs.append(
                    cx_fn(delayed_df)
                )
            else:
                delayed_dfs.append(delayed_df)

        return dd.from_delayed(delayed_dfs, meta=ddf._meta, divisions=ddf.divisions)


class _DaskPartitionCoordinateIndexer(_BaseCoordinateIndexer):
    def __init__(self, obj, sindex):
        super().__init__(sindex)
        self._obj = obj

    def _perform_get_item(self, covers_inds, overlaps_inds, x0, x1, y0, y1):
        covers_inds = set(covers_inds)
        overlaps_inds = set(overlaps_inds)
        all_partition_inds = sorted(covers_inds.union(overlaps_inds))
        if len(all_partition_inds) == 0:
            # No partitions intersect with query region, return empty result
            return dd.from_pandas(self._obj._meta, npartitions=1)

        return self._obj.partitions[all_partition_inds]
